
public class E.PantheonOnlineAccounts : E.Extension, E.OAuth2Support {
    Ag.Manager manager;
    HashTable<uint, string> uoa_to_eds;

    static construct {
        extensible_type = typeof (E.SourceRegistryServer);
    }

    construct {
        get_extensible ().bus_acquired.connect (() => {on_bus_acquired ();});
    }

    private void on_bus_acquired () {
        manager = new Ag.Manager ();
        uoa_to_eds = new HashTable<uint, string> (null, null);
        populate_accounts_table (manager.list ());
        foreach (var account_id in manager.list ()) {
            on_account_created (account_id);
        }

        manager.account_created.connect ((id) => {on_account_created (id);});
        manager.account_deleted.connect ((id) => {on_account_deleted (id);});
    }

    private void on_account_created (uint account_id) {
        var account = manager.get_account (account_id);
        var backend_name = POAUtils.get_backend_name (account.get_provider_name ());
        var source_uid = uoa_to_eds.get (account_id);

        if (source_uid == null && backend_name != null) {
            var backend_factory = ((E.DataFactory)get_extensible ()).ref_backend_factory (backend_name);
            if (backend_factory != null) {
                create_collection (account, backend_factory);
            }
        }
    }

    private void populate_accounts_table (List<uint> accounts_ids) {
        var trash = new Queue<E.Source> ();
        foreach (var source in ((E.SourceRegistryServer)get_extensible ()).list_sources (E.SOURCE_EXTENSION_UOA)) {
            var uoa_ext = (E.SourceUoa)source.get_extension (E.SOURCE_EXTENSION_UOA);
            var account_id = uoa_ext.get_account_id ();
            if (account_id == 0)
                continue;

            /* Verify the POA account still exists. */
            var match = accounts_ids.find (account_id).copy ();
            var account = manager.get_account (account_id);
            if (match != null && account != null) {
                uoa_to_eds.insert (account_id, source.get_uid ());
            } else {
                trash.push_tail (source);
            }
        }

        while (trash.is_empty () == false) {
            var source = trash.pop_head ();
            try {
                source.remove_sync ();
            } catch (Error e) {
                critical (e.message);
            }
        }
    }

    private void create_collection (Ag.Account account, E.BackendFactory backend_factory) {
        bool supports_oauth2 = false;
        var collection_source = new_source ();
        var account_services = new HashTable<string, Ag.AccountService> (GLib.str_hash, GLib.str_equal);

        /* Populate the hash table with AgAccountService instances by
         * service type. There should only be one AgService per type.*/
        foreach (var service in account.list_services ()) {
            account_services.insert (service.get_service_type (), new Ag.AccountService (account, service));
        }

        account.bind_property ("display-name", collection_source, "display-name", BindingFlags.SYNC_CREATE);
        collection_source.set_display_name (account.get_display_name ());
        account.bind_property ("enabled", collection_source, "enabled", BindingFlags.SYNC_CREATE);

        var source_extension = collection_source.get_extension (E.SOURCE_EXTENSION_UOA);
        account.bind_property ("id", source_extension, "account-id", BindingFlags.SYNC_CREATE);

        source_extension = collection_source.get_extension (E.SOURCE_EXTENSION_COLLECTION);
        account.bind_property ("provider", source_extension, "backend-name", BindingFlags.SYNC_CREATE, (binding, srcval, ref targetval) => {
            string src = (string) srcval;
            targetval.set_string (POAUtils.get_backend_name (src));
            return true;
        });

        var user_identity = account.get_display_name ();
        if (user_identity != null) {
            ((E.SourceCollection)source_extension).set_identity (user_identity);
        }

        var account_service = account_services.lookup (AG_SERVICE_TYPE_CALENDAR);
        if (account_service != null) {
            account_service.bind_property ("enabled", source_extension, "calendar-enabled", BindingFlags.SYNC_CREATE);
            supports_oauth2 |= POAUtils.check_oauth2_support (account_service);
        }

        account_service = account_services.lookup (AG_SERVICE_TYPE_CONTACTS);
        if (account_service != null) {
            account_service.bind_property ("enabled", source_extension, "contacts-enabled", BindingFlags.SYNC_CREATE);
            supports_oauth2 |= POAUtils.check_oauth2_support (account_service);
        }

        /* Stash the AgAccountService hash table in the ESource
         * to keep the property bindings alive.  The hash table
         * will be destroyed along with the ESource. */
        collection_source.set_data_full ("ag-account-services", account_services, null);

        /* The data source should not be removable by clients. */
        var server_side_source = (E.ServerSideSource)collection_source;
        server_side_source.set_removable (false);

        /* Obtain passwords from the signond service. */
        server_side_source.set_auth_session_type (typeof (GSignonSessionOAuth));

        if (supports_oauth2 == true) {
            /* This module provides OAuth 2.0 support to the collection.
             * Note, children of the collection source will automatically
             * inherit our EOAuth2Support through the property binding in
             * collection_backend_child_added(). */
            server_side_source.set_oauth2_support (this);
        }
        ((E.SourceRegistryServer)get_extensible ()).add_source (collection_source);

        uoa_to_eds.insert (account.id, collection_source.get_uid ());
    }

    private E.Source? new_source () {
        var file = E.ServerSideSource.new_user_file (null);
        E.Source source = null;
        try {
            source = new E.ServerSideSource ((E.SourceRegistryServer)get_extensible (), file);
        } catch (Error e) {
            critical (e.message);
        }

        return source;
    }

    private void on_account_deleted (uint account_id) {
        var source_uid = uoa_to_eds.lookup (account_id);
        if (source_uid == null)
            return;

        var source = ((E.SourceRegistryServer)get_extensible ()).ref_source (source_uid);
        if (source != null) {
            try {
                source.remove_sync ();
            } catch (Error e) {
                critical (e.message);
            }
        }
    }

    public bool get_access_token_sync (E.Source source, GLib.Cancellable? cancellable, out string access_token = null, out int expires_in = null) throws GLib.Error {
        var account_service = ref_account_service (source);
        if (account_service == null) {
            critical (_("Cannot find a corresponding account service in the accounts database from which to obtain an access token for '%s'"), source.get_display_name ());
            return false;
        }

        /* XXX This should never happen.  But because libaccounts-glib
         *     splits authentication method by service-type instead of
         *     by provider, and because we broadcast OAuth 2.0 support
         *     across the entire collection (spanning multiple service
         *     types), it's conceivable that not all service-types for
         *     a provider use OAuth 2.0, and an ESource for one of the
         *     ones that DOESN'T could mistakenly request the token. */
        if (!POAUtils.check_oauth2_support (account_service)) {
            critical (_("Data source '%s' does not support OAuth 2.0 authentication"), source.get_display_name ());
            return false;
        }

        var auth_data = account_service.get_auth_data ();
        var identity = new Signon.Identity.from_db (auth_data.get_credentials_id ());

        try {
            var session = identity.create_session (auth_data.get_method ());
            if (session != null) {
                var loop = new MainLoop ();
                Variant session_data = null;
                session.process_async.begin (auth_data.get_login_parameters ((GLib.Variant)null), auth_data.get_mechanism (), cancellable, (obj, res) => {
                    session.process_async.end (res);
                    session_data = (Variant)res;

                    loop.quit ();
                });
                loop.run ();
                if (session_data != null) {
                    session_data.lookup ("AccessToken", "s", out access_token);
                    session_data.lookup ("ExpiresIn", "i", out expires_in);
                }
            }
        } catch (Error e) {
            throw (e);
        }
        return false;
    }

    private Ag.AccountService? ref_account_service (E.Source source) {
        var uoasource = ((E.SourceRegistryServer)get_extensible ()).find_extension (source, E.SOURCE_EXTENSION_UOA);
        if (uoasource != null) {
            HashTable<string, Ag.AccountService> account_services = uoasource.get_data ("ag-account-services");
            if (account_services != null) {
                var service_type = POAUtils.get_service_type_from_source (uoasource);
                return account_services.lookup (service_type);
            }
        }

        return null;
    }

    public async bool get_access_token (E.Source source, GLib.Cancellable? cancellable, out string access_token = null, out int expires_in = null) throws GLib.Error {
        Signon.AuthSession session;
        var account_service = ref_account_service (source);
        if (account_service == null) {
            critical (_("Cannot find a corresponding account service in the accounts database from which to obtain an access token for '%s'"), source.get_display_name ());
            return false;
        }

        /* XXX This should never happen.  But because libaccounts-glib
         *     splits authentication method by service-type instead of
         *     by provider, and because we broadcast OAuth 2.0 support
         *     across the entire collection (spanning multiple service
         *     types), it's conceivable that not all service-types for
         *     a provider use OAuth 2.0, and an ESource for one of the
         *     ones that DOESN'T could mistakenly request the token. */
        if (!POAUtils.check_oauth2_support (account_service)) {
            critical (_("Data source '%s' does not support OAuth 2.0 authentication"), source.get_display_name ());
            return false;
        }

        var auth_data = account_service.get_auth_data ();
        var identity = new Signon.Identity.from_db (auth_data.get_credentials_id ());

        try {
            session = identity.create_session (auth_data.get_method ());
            if (session != null) {
                var session_data = yield session.process_async (auth_data.get_login_parameters ((GLib.Variant)null), auth_data.get_mechanism (), cancellable);
                if (session_data != null) {
                    session_data.lookup ("AccessToken", "s", out access_token);
                    session_data.lookup ("ExpiresIn", "i", out expires_in);
                }
            }
        } catch (Error e) {
            throw (e);
        }

        return false;
    }
}

[ModuleInit]
public static Type e_module_load (GLib.TypeModule type_module) {
    e_fake_module_load (type_module);
    return typeof(E.PantheonOnlineAccounts);
}

public void e_module_unload (TypeModule type_module) {
    
}